/* MoarNES source - main-haiku.cpp

   This open-source software is (c)2010-2013 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file is just a quick hackjob to get the MoarNES drop-down GUI menus
   working with Haiku's API. It's pretty ugly, but it works.
*/

#include "main-haiku.h"

#include <stdio.h>
#include <MenuBar.h>
#include <Menu.h>
#include <MenuItem.h>
#include <FilePanel.h>
#include <Entry.h>
#include <Path.h>
#include <View.h>

#include "config.h"

extern "C" {
	void setexitapp();
	void saveconfig();
	void setscreenmode(unsigned short modeval);
	void loadnewROM(char *romfile, char *fileonly);
	void closerom();
	void setgenie(unsigned char val);
	void configureinput(int inputval);
	void setCPUrun();
	void setCPUpause();
	void reset6502();
	void extloadstate(char *loadname);
	void extsavestate(char *savename);
	void resetAPU();
	void entry();
	void setexitapp();
	void mainloop();
}

void startemu(void *data) {
	mainloop();
}

int main(void) {
	status_t thread_result;
	
	entry();
	MainWindow *mainwin = new MainWindow();
	mainwin->Show();
	thread_id emuthread = spawn_thread((status_t (*)(void *))startemu, "", B_NORMAL_PRIORITY, NULL);
	wait_for_thread(emuthread, &thread_result);

	return 0;
}

MainWindow::MainWindow(void)
	:	BWindow(BRect(50,50,450,65),"MoarNES " BUILD_STRING,
			B_TITLED_WINDOW, B_ASYNCHRONOUS_CONTROLS) // | B_QUIT_ON_WINDOW_CLOSE)
{
	BRect r(Bounds());
	r.bottom = 20;
	
	BMenuBar *menuBar = new BMenuBar(r,"menubar");
	AddChild(menuBar);
	
	BMenu *filemenu = new BMenu("File");
	filemenu->AddItem(new BMenuItem("Open ROM...",new BMessage(M_FILE_OPEN_ROM)));
	filemenu->AddItem(new BMenuItem("Open ROM with Game Genie...",new BMessage(M_FILE_OPEN_GENIE)));
	filemenu->AddItem(new BMenuItem("Close ROM",new BMessage(M_FILE_CLOSE_ROM)));
	filemenu->AddSeparatorItem();
	//filemenu->AddItem(new BMenuItem("ROM information...",new BMessage(M_FILE_ROM_INFO),'K'));
	filemenu->AddItem(new BMenuItem("Load state...",new BMessage(M_FILE_LOAD_STATE)));
	filemenu->AddItem(new BMenuItem("Save state...",new BMessage(M_FILE_SAVE_STATE)));
	filemenu->AddSeparatorItem();
	filemenu->AddItem(new BMenuItem("Exit",new BMessage(M_FILE_EXIT)));
	menuBar->AddItem(filemenu);

	BMenu *emumenu = new BMenu("Emulation");
	menuBar->AddItem(emumenu);
	
	BMenu *cpumenu = new BMenu("CPU");
	cpumenu->AddItem(new BMenuItem("Run",new BMessage(M_EMU_RUN)));
	cpumenu->AddItem(new BMenuItem("Pause",new BMessage(M_EMU_PAUSE)));
	cpumenu->AddItem(new BMenuItem("Reset",new BMessage(M_EMU_RESET)));
	emumenu->AddItem(cpumenu);
	
	BMenu *inputmenu = new BMenu("Input");
	emumenu->AddItem(inputmenu);
	BMenu *pad1menu = new BMenu("Gamepad 1");
	inputmenu->AddItem(pad1menu);
	BMenu *pad2menu = new BMenu("Gamepad 2");
	inputmenu->AddItem(pad2menu);
	pad1menu->AddItem(new BMenuItem("Configure keyboard...",new BMessage(M_EMU_KEY_1)));
	pad1menu->AddItem(new BMenuItem("Configure joystick...",new BMessage(M_EMU_JOY_1)));
	pad2menu->AddItem(new BMenuItem("Configure keyboard...",new BMessage(M_EMU_KEY_2)));
	pad2menu->AddItem(new BMenuItem("Configure joystick...",new BMessage(M_EMU_JOY_2)));

	BMenu *videomenu = new BMenu("Video");
	menuBar->AddItem(videomenu);
	
	BMenu *zoommenu = new BMenu("Zoom");
	zoommenu->AddItem(new BMenuItem("1x",new BMessage(M_ZOOM_1X)));
	zoommenu->AddItem(new BMenuItem("2x",new BMessage(M_ZOOM_2X)));
	zoommenu->AddItem(new BMenuItem("3x",new BMessage(M_ZOOM_3X)));
	zoommenu->AddItem(new BMenuItem("4x",new BMessage(M_ZOOM_4X)));
	zoommenu->AddItem(new BMenuItem("Full screen",new BMessage(M_ZOOM_FULL)));
	videomenu->AddItem(zoommenu);
	
	//BView *view = new BView(BRect(100,100,300,200),"colorview",B_FOLLOW_ALL, B_WILL_DRAW);
	//AddChild(view);
	//view->SetViewColor(0,0,160);
	
	//entry();
}

BFilePanel *openPanel = NULL;
int panelresult;

void MainWindow::MessageReceived(BMessage *msg)
{
	//BView *view = FindView("colorview");
	
	switch (msg->what)
	{
		case M_FILE_OPEN_ROM:
		case M_FILE_OPEN_GENIE:
		{
			panelresult = PR_LOAD_ROM;
			if (msg->what == M_FILE_OPEN_GENIE) setgenie(1);
				else setgenie(0);
			ShowOpen();
			break;
		}
		case M_FILE_CLOSE_ROM:
		{
			closerom();
			resetAPU();
			break;
		}
		case M_FILE_LOAD_STATE:
		{
			panelresult = PR_LOAD_STATE;
			ShowOpen();
			break;
		}
		case M_FILE_SAVE_STATE:
		{
			panelresult = PR_SAVE_STATE;
			ShowSave();
			break;
		}
		case M_FILE_EXIT:
		{
//			setexitapp();
			saveconfig();
			exit(0);
			break;
		}
		case M_ZOOM_1X: setscreenmode(1); break;
		case M_ZOOM_2X: setscreenmode(2); break;
		case M_ZOOM_3X: setscreenmode(3); break;
		case M_ZOOM_4X: setscreenmode(4); break;
		case M_ZOOM_FULL: setscreenmode(5); break;
		case M_EMU_RUN: setCPUrun(); break;
		case M_EMU_PAUSE: setCPUpause(); break;
		case M_EMU_RESET: reset6502(); break;
		case M_EMU_KEY_1: configureinput(1); break;
		case M_EMU_JOY_1: configureinput(2); break;
		case M_EMU_KEY_2: configureinput(3); break;
		case M_EMU_JOY_2: configureinput(4); break;
		case B_REFS_RECEIVED:
		{
			entry_ref ref;
			if (msg->FindRef("refs", &ref) == B_OK) {
				BEntry entry(&ref);
				BPath path;
				if (entry.InitCheck() == B_OK) {
					entry.GetPath(&path);
					//printf("Path: %s\n", path.Path());
					switch (panelresult) {
						case PR_LOAD_ROM:
							closerom();
							resetAPU();
							loadnewROM((char *)path.Path(), ref.name);
							break;
						case PR_LOAD_STATE:
							extloadstate((char *)path.Path());
							break;
					}
				}
			}
			openPanel->~BFilePanel();
			break;
		}
		case B_SAVE_REQUESTED:
		{
			entry_ref ref;
			char fullsave[4096];
			const char *name = NULL;
			if (msg->FindRef("directory", &ref) == B_OK) {
				BEntry entry(&ref);
				BPath path;
				if (entry.InitCheck() == B_OK) {
					msg->FindString("name", &name);
					entry.GetPath(&path);
					sprintf(fullsave, "%s/%s", path.Path(), name);
					switch(panelresult) {
						case PR_SAVE_STATE:
							extsavestate(fullsave);
							break;
					}
				}
			}
			openPanel->~BFilePanel();
			break;
		}
		case B_QUIT_REQUESTED:
		{
			saveconfig();
			exit(0);
		}
		default:
		{
			//printf("Message received\n");
			BWindow::MessageReceived(msg);
			break;
		}
	}
}
