/* MoarNES source - menus.c

   This open-source software is (c)2010-2013 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains the code to handle the Win32 menubar that we throw on top
   of the SDL window via some Win32 API calls in render.c.
*/

#include <stdio.h>
#include "moarnes.h"

uint8_t msgdata[1024];

//extern uint8_t dodump;
extern uint8_t romloaded;
extern uint8_t rompath[MAX_PATH];
extern SDL_Surface *screen;
extern uint16_t changescreenmode;
extern const uint8_t *build;
extern bool running;
extern bool ROMwaiting;
extern bool paused;
extern bool exitapp;
extern uint32_t cursamplerate;
extern uint8_t nextROM[512];
extern uint8_t allowsweeps;
extern uint8_t savestatefilename[MAX_PATH];

extern bool doloadstate;
extern bool dosavestate;
extern uint8_t gamegenie;
extern uint8_t joycount;
extern struct cart_s cartridge;
extern void timedtext(uint8_t *text);
extern uint8_t outputNES[240][256];

extern uint8_t buf[48000];
extern int32_t bufferpos;

void pauseCPU() {
	if (paused == false) {
		SDL_PauseAudio(1);
		timedtext("CPU paused.");
		drawtext("CPU paused.", 0, 232);
		blitframe((uint8_t *)outputNES);
	}
	paused = true;
}

void resumeCPU() {
	if (paused == true) {
		timedtext("CPU resumed.");
		SDL_PauseAudio(0);
	}
	paused = false;
}

#ifdef _WIN32
void MenuItemClick(WPARAM wParam) {
	OPENFILENAME of_dlg;
	uint8_t filename[MAX_PATH] = { 0 };
	uint8_t textdata[256];

	switch (LOWORD(wParam)) {
		//file menu
		case ID_FILE_OPENROM:
		case ID_FILE_OPENROMWITHGAMEGENIE:
			if (running == true) {
				msgbox("You must close the current ROM before loading a new one.", MB_ICONINFORMATION | MB_OK);
				break;
			}
			memset(&of_dlg, 0, sizeof(of_dlg));
			of_dlg.lStructSize = sizeof(of_dlg);
			of_dlg.lpstrTitle = "Open ROM";
			of_dlg.hInstance = NULL;
			of_dlg.lpstrFile = filename;
			of_dlg.lpstrFilter = "NES ROM images (*.nes)\0*.nes\0\0";
			of_dlg.nMaxFile = MAX_PATH;
			of_dlg.Flags = OFN_FILEMUSTEXIST | OFN_LONGNAMES;
			if (GetOpenFileName(&of_dlg)) {
				if (LOWORD(wParam) == ID_FILE_OPENROMWITHGAMEGENIE) gamegenie = 1;
					else gamegenie = 0;
				if (loadROM((uint8_t *)of_dlg.lpstrFile) == false) return;
				sprintf((int8_t *)nextROM, of_dlg.lpstrFile);
				running = false;
				ROMwaiting = true;
				timedtext("ROM loaded.");
			}
			break;

		case ID_FILE_CLOSEROM:
			SDL_PauseAudio(1);
			memset(buf, 0, 48000);
			bufferpos = 0;
			running = false;
			ROMwaiting = false;
			break;

		case ID_FILE_ROMINFORMATION:
			if (!romloaded) {
				msgbox("No ROM is currently loaded.", MB_ICONINFORMATION | MB_OK);
			} else {
				sprintf((int8_t *)textdata, "ROM information:\n\nFile: %s\nMapper: %u\nPRG data: %u KB\nCHR data: %u KB", rompath, cartridge.mapper, cartridge.PRGcount, cartridge.CHRcount);
				msgbox(textdata, MB_ICONINFORMATION | MB_OK);
			}
			break;
		
		case ID_FILE_LOADSTATE:
			memset(&of_dlg, 0, sizeof(of_dlg));
			of_dlg.lStructSize = sizeof(of_dlg);
			of_dlg.lpstrTitle = "Open state";
			of_dlg.hInstance = NULL;
			of_dlg.lpstrFile = (LPSTR)filename;
			of_dlg.lpstrFilter = "MoarNES save states (*.mns)\0*.mns\0\0";
			of_dlg.nMaxFile = MAX_PATH;
			of_dlg.Flags = OFN_FILEMUSTEXIST | OFN_LONGNAMES;
			if (GetOpenFileName(&of_dlg)) {
				sprintf((int8_t *)savestatefilename, of_dlg.lpstrFile);
				doloadstate = true;
			}
			break;

		case ID_FILE_SAVESTATE:
			memset(&of_dlg, 0, sizeof(of_dlg));
			of_dlg.lStructSize = sizeof(of_dlg);
			of_dlg.lpstrTitle = "Save state";
			of_dlg.hInstance = NULL;
			of_dlg.lpstrFile = (LPSTR)filename;
			of_dlg.lpstrFilter = "MoarNES save states (*.mns)\0*.mns\0\0";
			of_dlg.nMaxFile = MAX_PATH;
			of_dlg.Flags = OFN_LONGNAMES;
			if (GetSaveFileName(&of_dlg)) {
				sprintf((int8_t *)savestatefilename, "%s.mns", of_dlg.lpstrFile);
				if (memcmp(&savestatefilename[strlen((int8_t *)savestatefilename) - 8], ".mns.mns", 4) == 0) {
					savestatefilename[strlen((int8_t *)savestatefilename) - 4] = 0;
				}
				dosavestate = true;
			}
			break;

		case ID_FILE_EXIT:
			verifyexit();
			break;

		//emulation menu
		case ID_CPU_RUN:
			resumeCPU();
			break;

		case ID_CPU_PAUSE:
			pauseCPU();
			break;

		case ID_CPU_RESET:
			reset6502();
			break;

		case ID_GAMEPAD1_CONFIGUREKEYBOARD:
			configurekeyboard(0);
			break;

		case ID_GAMEPAD1_CONFIGUREJOYSTICK:
			if (joycount) configurejoystick(0);
				else msgbox("There are no joysticks connected to this computer!", MB_ICONINFORMATION | MB_OK);
			break;

		case ID_GAMEPAD2_CONFIGUREKEYBOARD:
			configurekeyboard(1);
			break;

		case ID_GAMEPAD2_CONFIGUREJOYSTICK:
			if (joycount) configurejoystick(1);
				else msgbox("There are no joysticks connected to this computer!", MB_ICONINFORMATION | MB_OK);
			break;

		//video menu
		case ID_ZOOM_1X:
			//setscreenmode(1);
			changescreenmode = 1;
			break;

		case ID_ZOOM_2X:
			//setscreenmode(2);
			changescreenmode = 2;
			break;

		case ID_ZOOM_3X:
			//setscreenmode(3);
			changescreenmode = 3;
			break;

		case ID_ZOOM_4X:
			//setscreenmode(4);
			changescreenmode = 4;
			break;

		case ID_ZOOM_FULLSCREEN:
			//setscreenmode(5);
			changescreenmode = 5;
			break;

		//audio menu
		case ID_SAMPLERATE_96000HZ:
			killAPU();
			if (initAPU(1, 96000) == false) {
				sprintf((int8_t *)msgdata, "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(msgdata, MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		case ID_SAMPLERATE_48000HZ:
			killAPU();
			if (initAPU(1, 48000) == false) {
				sprintf((int8_t *)msgdata, "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(msgdata, MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		case ID_SAMPLERATE_44100HZ:
			killAPU();
			if (initAPU(1, 44100) == false) {
				sprintf((int8_t *)msgdata, "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(msgdata, MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		case ID_SAMPLERATE_32000HZ:
			killAPU();
			if (initAPU(1, 32000) == false) {
				sprintf((int8_t *)msgdata, "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(msgdata, MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		case ID_SAMPLERATE_22050HZ:
			killAPU();
			if (initAPU(1, 22050) == false) {
				sprintf((int8_t *)msgdata, "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(msgdata, MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		case ID_SAMPLERATE_11025HZ:
			killAPU();
			if (initAPU(1, 11025) == false) {
				sprintf((int8_t *)msgdata, "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(msgdata, MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		/*case ID_AUDIO_ENABLESWEEPUNITS:
			allowsweeps = 1;
			break;

		case ID_AUDIO_DISABLESWEEPUNITS:
			allowsweeps = 0;
			break;*/

        //help menu
		case ID_HELP_ABOUT:
			showabout();
			break;
	}
}

void showabout() {
	uint8_t about[512];

	sprintf((int8_t *)about, "MoarNES %s (c)2010-2013 Mike Chambers\nMoarNES is distributed under the terms of the GNU GPL v2.\n\nHome page: http://sourceforge.net/projects/moarnes", build);
	msgbox(about, MB_ICONINFORMATION | MB_OK);
}
#endif
