/* MoarNES source - render.c

   This open-source software is (c)2010-2013 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains the code to render NES frames to the screen, as well
   as managing the SDL screen context.

   Changelog:
     v0.12.2.22 alpha - Fixed PPU read buffering for PPU memory below $2000.
	                    Super Mario Bros. title screen now looks correct.
*/

#include <stdio.h>
#include "common.h"
#ifdef _WIN32
	#include "SDL/SDL_syswm.h"
#endif
#include "moarnes.h"
#include "font.h"

const uint32_t paletteNES[66] = {
   0x808080, 0x0000BB, 0x3700BF, 0x8400A6,
   0xBB006A, 0xB7001E, 0xB30000, 0x912600,
   0x7B2B00, 0x003E00, 0x00480D, 0x003C22,
   0x002F66, 0x000000, 0x050505, 0x050505,

   0xC8C8C8, 0x0059FF, 0x443CFF, 0xB733CC,
   0xFF33AA, 0xFF375E, 0xFF371A, 0xD54B00,
   0xC46200, 0x3C7B00, 0x1E8415, 0x009566,
   0x0084C4, 0x111111, 0x090909, 0x090909,

   0xFFFFFF, 0x0095FF, 0x6F84FF, 0xD56FFF,
   0xFF77CC, 0xFF6F99, 0xFF7B59, 0xFF915F,
   0xFFA233, 0xA6BF00, 0x51D96A, 0x4DD5AE,
   0x00D9FF, 0x666666, 0x0D0D0D, 0x0D0D0D,

   0xFFFFFF, 0x84BFFF, 0xBBBBFF, 0xD0BBFF,
   0xFFBFEA, 0xFFBFCC, 0xFFC4B7, 0xFFCCAE,
   0xFFD9A2, 0xCCE199, 0xAEEEB7, 0xAAF7EE,
   0xB3EEFF, 0xDDDDDD, 0x111111, 0x111111, 0xFFFFFF, 0x000000};

SDL_Surface *screen;
uint16_t screenmode = 2, lastscreenmode = 2, changescreenmode = 0;

extern uint8_t outputNES[240][256];

#ifdef _WIN32
HWND myWindow;
HINSTANCE myInstance;
HMENU myMenu;
WNDPROC oldProc;
HWND GetHwnd();
HICON myIcon;
void SetWndProc();
#endif

extern uint32_t totalframes;

bool initSDL(uint8_t *windowcaption) {
	if (SDL_Init(SDL_INIT_EVERYTHING) < 0) return(false);

	screen = SDL_SetVideoMode(512, 480, 32, SDL_HWSURFACE);
	if (screen == NULL) {
		SDL_Quit();
		return(false);
	}

	SDL_WM_SetCaption(windowcaption, NULL);
	#ifdef _WIN32
		myWindow = GetHwnd();
		myInstance = GetModuleHandle(NULL);
		myMenu = LoadMenu(NULL, MAKEINTRESOURCE(IDR_MENU1));
		SetMenu(myWindow, myMenu);
		SetWndProc();
		screen = SDL_SetVideoMode(512, 480, 32, SDL_HWSURFACE);
		if (screen == NULL) {
			SDL_Quit();
			return(false);
		}
		myIcon = LoadIcon(myInstance, MAKEINTRESOURCE(IDI_ICON1));
		SetClassLong(myWindow, GCLP_HICON, (LONG)(uint64_t)myIcon);
		SetClassLong(myWindow, GCLP_HICONSM, (LONG)(uint64_t)myIcon);
	#endif

	return(true);
}

int8_t *readSDLerr() {
	return(SDL_GetError());
}

#ifdef _WIN32
HWND GetHwnd() {
	SDL_SysWMinfo wmi;
	SDL_VERSION(&wmi.version);

	if (!SDL_GetWMInfo(&wmi)) return(NULL);
	return(wmi.window);
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch (msg) {
		case WM_COMMAND:
			MenuItemClick(wParam);
			return(TRUE);
	}

	return(CallWindowProc(oldProc, hwnd, msg, wParam, lParam));
}

void SetWndProc() {
	oldProc = (WNDPROC)SetWindowLong(myWindow, GWL_WNDPROC, (LONG_PTR)WndProc);
}
#endif

void drawtext(uint8_t *text, uint16_t x, uint16_t y) {
	uint8_t curchar = 0;
	uint16_t textpos = 0, calcx, calcy;

	for (;;) {
		curchar = text[textpos++];
		if (curchar == 0) break;
		for (calcy=y; calcy<(y+8); calcy++) {
			for (calcx=x; calcx<(x+8); calcx++) {
				outputNES[calcy][calcx] = 65 - font[curchar][(calcy-y)*8+(calcx-x)];
			}
		}
		x += 8;
	}
}

bool blitting = false;
void blitframe(uint8_t *NESbuffer) {
	uint32_t bufptr, SDLptr = 0;
	int32_t x, y, calcx, calcy;

	blitting = true;

	if (SDL_MUSTLOCK(screen)) 
		if (SDL_LockSurface(screen) < 0) 
			return;

	switch (screenmode) {
		case 1:
			bufptr = 0;
			SDLptr = 0;
			for (y=0; y<screen->h; y++) {
				for (x=0; x<screen->w; x++) {
					((uint32_t *)screen->pixels)[SDLptr++] = paletteNES[NESbuffer[bufptr++]];
				}
			}
			break;

		case 2:
			bufptr = 0;
			SDLptr = 0;
			for (y=0; y<480; y+=2) {
				for (x=0; x<512; x++) {
					((uint32_t *)screen->pixels)[SDLptr] = paletteNES[NESbuffer[bufptr]];
					((uint32_t *)screen->pixels)[SDLptr++ + screen->w] = paletteNES[NESbuffer[bufptr]];
					if (x & 1) bufptr++;
				}
				SDLptr += screen->w;
			}
			break;

		case 3:
			for (y=0; y<screen->h; y++) {
				for (x=0; x<screen->w; x++) {
					calcx = x / 3;
					calcy = y / 3;
					bufptr = (calcy << 8) + calcx;
					SDLptr = (y * screen->w) + x;
					((uint32_t *)screen->pixels)[SDLptr] = paletteNES[NESbuffer[bufptr]];
				}
			}
			break;

		case 4:
			for (y=0; y<screen->h; y++) {
				for (x=0; x<screen->w; x++) {
					calcx = x >> 2;
					calcy = y >> 2;
					bufptr = (calcy << 8) + calcx;
					SDLptr = (y * screen->w) + x;
					((uint32_t *)screen->pixels)[SDLptr] = paletteNES[NESbuffer[bufptr]];
				}

			}
			break;

		case 5:
			bufptr = 0;
			if (screen->h == 240) {
				for (y=0; y<240; y++) {
					SDLptr = (y * screen->w) + 32;
					for (x=0; x<256; x++) {
						((uint32_t *)screen->pixels)[SDLptr++] = paletteNES[NESbuffer[bufptr++]];
					}
				}
				break;
			} else {
				for (y=0; y<480; y+=2) {
					SDLptr = (y * screen->w) + 64;
					for (x=0; x<512; x++) {
						((uint32_t *)screen->pixels)[SDLptr] = paletteNES[NESbuffer[bufptr]];
						((uint32_t *)screen->pixels)[SDLptr++ + screen->w] = paletteNES[NESbuffer[bufptr]];
						if (x & 1) bufptr++;
					}
				}
				break;
			}
	}

  if (SDL_MUSTLOCK(screen)) 
    SDL_UnlockSurface(screen);

  SDL_UpdateRect(screen, 0, 0, screen->w, screen->h);
  blitting = false;
}

void setscreenmode(uint16_t modeval) {
	switch (modeval) {
		case 1:
			screen = SDL_SetVideoMode(256, 240, 32, SDL_HWSURFACE);
#ifdef _WIN32
			if (screenmode == 5) SetMenu(myWindow, myMenu);
#endif
			lastscreenmode = screenmode;
			screenmode = 1;
			break;

		case 2:
			screen = SDL_SetVideoMode(512, 480, 32, SDL_HWSURFACE);
#ifdef _WIN32
			if (screenmode == 5) SetMenu(myWindow, myMenu);
#endif
			lastscreenmode = screenmode;
			screenmode = 2;
			break;

		case 3:
			screen = SDL_SetVideoMode(768, 720, 32, SDL_HWSURFACE);
#ifdef _WIN32
			if (screenmode == 5) SetMenu(myWindow, myMenu);
#endif
			lastscreenmode = screenmode;
			screenmode = 3;
			break;

		case 4:
			screen = SDL_SetVideoMode(1024, 960, 32, SDL_HWSURFACE);
#ifdef _WIN32
			if (screenmode == 5) SetMenu(myWindow, myMenu);
#endif
			lastscreenmode = screenmode;
			screenmode = 4;
			break;

		case 5:
#ifdef _WIN32
			SetMenu(myWindow, NULL);
#endif
			/*screen = SDL_SetVideoMode(320, 240, 32, SDL_HWSURFACE | SDL_FULLSCREEN);
			if (screen->h != 240)*/ screen = SDL_SetVideoMode(640, 480, 32, SDL_HWSURFACE | SDL_FULLSCREEN);
			lastscreenmode = screenmode;
			screenmode = 5;
			break;
	}

	if (modeval == 5) SDL_ShowCursor(SDL_DISABLE);
		else SDL_ShowCursor(SDL_ENABLE);
	blitframe(&outputNES[0][0]);
}
