/* MoarNES source - rom.c

   This open-source software is (c)2010-2013 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains the code to load NES ROM files.
*/

#include <stdio.h>
#include <malloc.h>
#include "moarnes.h"
#include "genie.h"

uint8_t gamegenie = 0;
uint8_t rompath[MAX_PATH];
uint8_t romloaded = 0;
extern bool ROMisNSF;
struct cart_s cartridge;
extern struct PPU_s *PPU;
static const uint8_t validhdr[4] = { 'N', 'E', 'S', 0x1A };
extern uint8_t app_path[MAX_PATH];

bool loadROM(uint8_t *filename) {
	FILE *romfile = NULL;
	static struct hdr_s header;
	uint16_t setptr;
	uint8_t path[MAX_PATH];

	if (gamegenie == 0) {
		romfile = fopen(filename, "rb");
		sprintf(&rompath[0], filename);
	} else if (gamegenie == 1) {
		sprintf(&rompath[0], filename);
	} else if (gamegenie == 2) {
		romfile = fopen(&rompath[0], "rb");
	}
	romloaded = 1;
	//if (ROMisNSF) goto NSFloader;

	if (gamegenie == 1) memcpy(&header, genieROM, sizeof(header));
		else fread(&header, 1, sizeof(header), romfile);
	if (memcmp((void *)&header.match, (void *)&validhdr[0], 4)) return(false);

	//printf("ROM information:\n");
	//parse through the header data
	if (header.flags[1] == 'D') header.flags[1] = 0; //DISKDUDE fix
	cartridge.mapper = ((header.flags[0] >> 4) | header.flags[1] & 0xF0);
	cartridge.hasSRAM = 1; //(header.flags[0] & 2) >> 1;
	if (header.flags[0] & 8) cartridge.mirroring = MIR_FOURSCREEN;
		else if (header.flags[0] & 1) cartridge.mirroring = MIR_VERTICAL;
			else cartridge.mirroring = MIR_HORIZONTAL;
	if (header.flags[0] & 4) cartridge.hasTrainer = true;
		else cartridge.hasTrainer = false;
	cartridge.PRGcount = (uint16_t)header.PRGsize << 4; //header gives PRG size in 16 KB units, we want it in 1 KB
	cartridge.CHRcount = (uint16_t)header.CHRsize << 3; //and CHR size is in 8 KB units

	switch (cartridge.mirroring) {
		case 0:
			PPU->ntmap[0] = 0x2000;
			PPU->ntmap[1] = 0x2000;
			PPU->ntmap[2] = 0x2400;
			PPU->ntmap[3] = 0x2400;
			break;
		case 1:
			PPU->ntmap[0] = 0x2000;
			PPU->ntmap[1] = 0x2400;
			PPU->ntmap[2] = 0x2000;
			PPU->ntmap[3] = 0x2400;
			break;
		case 2:
			PPU->ntmap[0] = 0x2000;
			PPU->ntmap[1] = 0x2400;
			PPU->ntmap[2] = 0x2600;
			PPU->ntmap[3] = 0x2800;
			break;
		default:
			break;
	}

	cartridge.PRGfull = malloc(cartridge.PRGcount << 10);
	if (cartridge.PRGfull == NULL) fatalerr("Unable to allocate for PRG-ROM data!");
	if (cartridge.CHRcount > 0) {
		cartridge.CHRfull = malloc(cartridge.CHRcount << 10);
		if (cartridge.CHRfull == NULL) fatalerr("Unable to allocate for CHR-ROM data!");
	} else {
		cartridge.CHRfull = malloc(1024 << 10);
	}
	if (gamegenie != 1) {
		if (cartridge.hasTrainer == true) fseek(romfile, 528, SEEK_SET);
			else fseek(romfile, 16, SEEK_SET);
		if (fread(cartridge.PRGfull, 1, (uint32_t)cartridge.PRGcount << 10, romfile) < ((uint32_t)cartridge.PRGcount << 10))
			fatalerr("The ROM appears to be missing PRG data!");
		if (cartridge.CHRcount > 0) if (fread(cartridge.CHRfull, 1, (uint32_t)cartridge.CHRcount << 10, romfile) < (uint32_t)cartridge.CHRcount << 10)
			fatalerr("The ROM appears to be missing CHR data!");	
		fclose(romfile);
	} else {
		memcpy(cartridge.PRGfull, &genieROM[16], 16384);
		memcpy(cartridge.CHRfull, &genieROM[16400], 8192);
	}

	//first init all PRG/CHR bank pointers to the beginning of PRG/CHR memory
	//just in case, to avoid possible crashes from bad pointers
	for (setptr=0; setptr<1024; setptr++) {
		cartridge.CHRbank[setptr] = cartridge.CHRfull;
		cartridge.PRGbank[setptr] = cartridge.PRGfull;
	}

	if (mapperinit(&cartridge) == false) return(false);

	//delete any existing quick save files
	sprintf(path, "%sslot1.mns", app_path); remove(path);
	sprintf(path, "%sslot2.mns", app_path); remove(path);
	sprintf(path, "%sslot3.mns", app_path); remove(path);
	sprintf(path, "%sslot4.mns", app_path); remove(path);
	return(true);
}
