/* MoarNES source - state.c

   This open-source software is (c)2010-2013 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains the code to load and save files containing gamestates,
   so that play may be resumed later from a specific point.
*/

#include <stdio.h>
#include "config.h"
#include "moarnes.h"

uint8_t savestatefilename[MAX_PATH];
extern bool dosavestate;
extern bool doloadstate;

extern int16_t msgbox(uint8_t *msgtxt, uint16_t buttons);

extern struct PPU_s *PPU;
extern struct OAM_s *OAM;
extern uint8_t *VRAM;
extern void *mappermemory;
extern uint32_t mappersize;
extern struct cart_s cartridge;
extern uint8_t NESRAM[0x800];

extern uint16_t pc;
extern uint8_t sp, a, x, y, status;

extern int8_t channels[5];
extern struct square_s square[2];
extern struct triangle_s triangle;
extern struct noise_s noise;
extern struct dmc_s dmc;

extern uint32_t seqtickgap;
extern uint8_t seqstep, seqmode, interruptAPU;

extern void timedtext(uint8_t *text);

void loadstate() {
	FILE *savefile = NULL;
	uint16_t bankrestore;

	doloadstate = false;

	savefile = fopen((int8_t *)savestatefilename, "rb");
	if (savefile == NULL) {
		timedtext("Load state failed.");
		return;
	}

	fread(&NESRAM[0], 1, 2048, savefile);
	fread(&cartridge.SRAM[0], 1, 8192, savefile);
	fread(PPU, 1, sizeof(struct PPU_s), savefile);
	fread(OAM, 1, sizeof(struct OAM_s), savefile);
	fread(VRAM, 1, 16384, savefile);
	fread(&mappersize, 1, 4, savefile);
	fread(mappermemory, 1, mappersize, savefile);
	fread(&cartridge.PRGblock[0], 1, 2048, savefile);
	fread(&cartridge.CHRblock[0], 1, 2048, savefile);
	fread(&pc, 1, 2, savefile);
	fread(&sp, 1, 1, savefile);
	fread(&a, 1, 1, savefile);
	fread(&x, 1, 1, savefile);
	fread(&y, 1, 1, savefile);
	fread(&status, 1, 1, savefile);
	fread(&channels[0], 1, 5, savefile);
	fread(&square[0], 1, sizeof(struct square_s), savefile);
	fread(&square[1], 1, sizeof(struct square_s), savefile);
	fread(&triangle, 1, sizeof(struct triangle_s), savefile);
	fread(&noise, 1, sizeof(struct noise_s), savefile);
	fread(&dmc, 1, sizeof(struct dmc_s), savefile);
	fread(&seqtickgap, 1, 4, savefile);
	fread(&seqstep, 1, 1, savefile);
	fread(&seqmode, 1, 1, savefile);
	fread(&interruptAPU, 1, 1, savefile);
	if (cartridge.CHRcount == 0)
		fread(cartridge.CHRfull, 1, 256 << 10, savefile);
	fclose(savefile);

	for (bankrestore=0; bankrestore<1024; bankrestore++) {
		cartridge.PRGbank[bankrestore] = &cartridge.PRGfull[(uint32_t)cartridge.PRGblock[bankrestore] << 10];
		cartridge.CHRbank[bankrestore] = &cartridge.CHRfull[(uint32_t)cartridge.CHRblock[bankrestore] << 10];
	}

	timedtext("Loaded save state.");
}

void savestate() {
	FILE *savefile = NULL;

	dosavestate = false;

	savefile = fopen((int8_t *)savestatefilename, "wb");
	if (savefile == NULL) {
		timedtext("Save state failed.");
		return;
	}

	fwrite(&NESRAM[0], 1, 2048, savefile);
	fwrite(&cartridge.SRAM[0], 1, 8192, savefile);
	fwrite(PPU, 1, sizeof(struct PPU_s), savefile);
	fwrite(OAM, 1, sizeof(struct OAM_s), savefile);
	fwrite(VRAM, 1, 16384, savefile);
	fwrite(&mappersize, 1, 4, savefile);
	fwrite(mappermemory, 1, mappersize, savefile);
	fwrite(&cartridge.PRGblock[0], 1, 2048, savefile);
	fwrite(&cartridge.CHRblock[0], 1, 2048, savefile);
	fwrite(&pc, 1, 2, savefile);
	fwrite(&sp, 1, 1, savefile);
	fwrite(&a, 1, 1, savefile);
	fwrite(&x, 1, 1, savefile);
	fwrite(&y, 1, 1, savefile);
	fwrite(&status, 1, 1, savefile);
	fwrite(&channels[0], 1, 5, savefile);
	fwrite(&square[0], 1, sizeof(struct square_s), savefile);
	fwrite(&square[1], 1, sizeof(struct square_s), savefile);
	fwrite(&triangle, 1, sizeof(struct triangle_s), savefile);
	fwrite(&noise, 1, sizeof(struct noise_s), savefile);
	fwrite(&dmc, 1, sizeof(struct dmc_s), savefile);
	fwrite(&seqtickgap, 1, 4, savefile);
	fwrite(&seqstep, 1, 1, savefile);
	fwrite(&seqmode, 1, 1, savefile);
	fwrite(&interruptAPU, 1, 1, savefile);
	if (cartridge.CHRcount == 0)
		fwrite(cartridge.CHRfull, 1, 256 << 10, savefile);
	fclose(savefile);

	timedtext("Saved state to file.");
}
